use rusqlite::{Connection};
use std::{sync::mpsc::{Sender, Receiver, channel, SendError}, thread::{self, JoinHandle}, any::Any};

enum DBMessage<T> {
    HangUp,
    Message(T)
}

#[derive(Debug)]
pub struct DBSender<MsgT> {
    sender: Sender<DBMessage<MsgT>>
}

// #[derive(Clone)] requires MsgT to be Clone, but that's not necessary
impl<MsgT> Clone for DBSender<MsgT> {
    fn clone(&self) -> Self {
        Self {sender: self.sender.clone()}
    }
}

impl<MsgT> DBSender<MsgT> {
    pub fn send(&self, msg: MsgT) -> Result<(), SendError<MsgT>>{
        if let Err(err) = self.sender.send(DBMessage::Message(msg)) {
            if let DBMessage::Message(msg) = err.0 {
                Err(SendError(msg))
            } else {
                unreachable!();
            }
        } else {
            Ok(())
        }
    }
    
    fn hang_up(&self) -> Result<(), SendError<()>> {
        if let Err(_) = self.sender.send(DBMessage::HangUp) {
            Err(SendError(()))
        }
        else {
            Ok(())
        }
    }
}


#[derive(Debug)]
pub struct DBConnection<MsgT: Send + 'static> 
{
    pub sender: DBSender<MsgT>,
    join_handle: Option<JoinHandle<()>>
}


impl<MsgT: Send + 'static> DBConnection<MsgT>
    {
    pub fn init<F: Fn(MsgT, &Connection) + Send + 'static>(connection: Connection, op: F) -> Self {
        let (sender, receiver) = channel();
    
        let join_handle = thread::spawn(move || Self::db_thread(receiver, connection, op));
        let sender = DBSender { sender };

        Self {sender, join_handle: Some(join_handle)}
    }
    
    fn db_thread<F: Fn(MsgT, &Connection) + Send>(receiver: Receiver<DBMessage<MsgT>>, connection: Connection, op: F) {
        while let Ok(db_msg) = receiver.recv() {
            // _ = connection.execute("INSERT INTO hits VALUES(?1, ?2, ?3, ?4);", params![hit.pixel_id, hit.ip, hit.user_agent, hit.date]).unwrap();
            if let DBMessage::Message(msg) = db_msg {
                op(msg, &connection);
            } 
            else if let DBMessage::HangUp = db_msg {
                break;
            }
        }
        
        // execute last messages in buffer
        for db_msg in receiver.try_iter() {
            if let DBMessage::Message(msg) = db_msg {
                op(msg, &connection);
            }
        }
    }
                                                        // static is implied, but included for clearness
    pub fn finish(&mut self) -> Result<(), Box<dyn Any + Send + 'static>> {
        // let Self{sender, join_handle} = self;
        if let Err(err) = self.sender.hang_up() {
            return Err(Box::new(err));
        }
        if let Some(handle) = self.join_handle.take() {
            handle.join()
        } 
        else {
            Ok(())
        }
    }
}

impl<MsgT: Send> Drop for DBConnection<MsgT> {
    fn drop(&mut self) {
        self.finish().unwrap();
    }
}