use std::{fs, sync::{Mutex}};

use askama::Template;
use rouille::{Response, router, input, Request, try_or_400, post_input};


use pixel::{Hit, PixelManager, PixelManagerDelegate, TrackingPixel, User};

mod db;
mod pixel;
mod helpers;


fn main() {
    let white_1x1_pixel_png = fs::read("assets/white_pixel.png").unwrap();

    let pixel_manager = PixelManager::build("app.db").unwrap();

    let pixel_manager_delegate = PixelManagerDelegate::new(&pixel_manager);
    // let pixel_id = pixel_manager.create_pixel("Hallo mac").unwrap();
    // print!("{pixel_id}");
    // pixel_manager.register_hit(pixel_id, "ai pie", "joeser eedsjent", chrono::offset::Utc::now());
    
    let delegate_mutex = Mutex::new(pixel_manager_delegate);
    
    rouille::start_server("0.0.0.0:3737", move |request| {
        router!(request,
            (GET) ["/manage"] => {
                let delegate = delegate_mutex.lock().unwrap().clone();
                // panicking on poisoned mutex is acceptable
                manage_index(request, &delegate)
            },
            (GET) ["/manage/pixel/{pixel_id}", pixel_id: String] => {
                let delegate = delegate_mutex.lock().unwrap().clone();
                manage_pixel(&pixel_id, request, &delegate)
            },
            (POST) ["/manage/pixel/create"] => {
                let delegate = delegate_mutex.lock().unwrap().clone();
                create_pixel(request, &delegate)
            },
            (POST) ["/manage/pixel/delete"] => {
                let delegate = delegate_mutex.lock().unwrap().clone();
                delete_pixel(request, &delegate)
            },
            (GET) ["/img/{id}/white.png", id: String] => {
                let delegate = delegate_mutex.lock().unwrap().clone();
                register_hit(&id, request, &delegate);
                Response::from_data("image/png", white_1x1_pixel_png.clone())
            },
            _ => {
                println!("404: {:#?}", request.url());
                Response::empty_404()
            }
        )
    });

}

#[derive(Template)]
#[template(path = "index.html")]
struct IndexTemplate<'a> {
    username: &'a str,
    pixels: Vec<(String, String)>
}

impl<'a> IndexTemplate<'a> {
    fn new(username: &'a str, pixels: &Vec<TrackingPixel>) -> Self {
        Self {
            username,
            pixels: pixels.iter().map(|pix| (pix.name.clone(), pix.url_safe_encode())).collect()
        }
    }
}


#[derive(Template)]
#[template(path = "manage/pixel.html")]
struct PixelTemplate {
    pixel_name: String,
    hits: Vec<(String, String, String)>,
    pixel_url: String,
    pixel_id: String
}

impl PixelTemplate {
    fn new(pixel: &TrackingPixel, hits: &Vec<Hit>) -> Self {
        Self {
            pixel_name: pixel.name.clone(),
            hits: hits.iter().map(|hit| (hit.date.to_string(), hit.ip.clone(), hit.user_agent.clone())).collect(),
            pixel_url: get_url_for_pixel(&pixel),
            pixel_id: pixel.url_safe_encode()
        }
    }
}

fn manage_index(request: &Request, manager_delegate: &PixelManagerDelegate) -> Response {
    let user = match auth_user(request, &manager_delegate) {
        Ok(user) => user,
        Err(response) => return response
    };
    
    match manager_delegate.get_manager().get_pixels_for_user(&user) {
        Ok(pixels) => {
            let index_template = IndexTemplate::new(&user.username, &pixels);
            match index_template.render() {
                Ok(page) => Response::html(page),
                Err(_) => Response::text("Page error").with_status_code(500)
            }
        },
        Err(_) => Response::text("DB error").with_status_code(500)
    }
}

fn auth_user(request: &Request, manager_delegate: &PixelManagerDelegate) -> Result<User, Response> {
    let auth = match input::basic_http_auth(request) {
        Some(a) => a,
        None => return Err(Response::basic_http_auth_login_required("manage"))
    };
    match manager_delegate.get_manager().get_user(&auth.login, &auth.password) {
        Ok(user) => Ok(user),
        Err(_) => return Err(Response::basic_http_auth_login_required("manage"))
    }
}

fn manage_pixel(pixel_id: &str, request: &Request, manager_delegate: &PixelManagerDelegate) -> Response {
    let user = match auth_user(request, &manager_delegate) {
        Ok(user) => user,
        Err(response) => return response
    };
    
    let pixel = match manager_delegate.get_manager().get_pixel_from_encoded(pixel_id, Some(&user)) {
        Ok(pixel_opt) => {
            match pixel_opt {
                Some(pixel) => pixel,
                None => return Response::text("Unknown pixel").with_status_code(500)
            }
        },
        Err(_) => return Response::text("Encoding error").with_status_code(500)
    };

    let hits = match manager_delegate.get_manager().get_hits_for_pixel(&pixel, Some(&user)) {
        Ok(hits) => hits,
        Err(_) => return Response::text("DB error").with_status_code(500)
    };

    let pixel_template = PixelTemplate::new(&pixel, &hits);

    match pixel_template.render() {
        Ok(page) => Response::html(page),
        Err(_) => Response::text("Page error").with_status_code(500)
    }
}

fn create_pixel(request: &Request, manager_delegate: &PixelManagerDelegate) -> Response {
    let user = match auth_user(request, &manager_delegate) {
        Ok(user) => user,
        Err(response) => return response
    };
    let input = try_or_400!(post_input!(request, {
        pixel_name: String
    }));
    let _pixel = match manager_delegate.get_manager().create_pixel(&input.pixel_name, &user) {
        Ok(pixel) => pixel,
        Err(_) => return Response::empty_400()
    };
    // Response::redirect_303("/manage")
    helpers::javascript_redirect("/manage")
}

fn delete_pixel(request: &Request, manager_delegate: &PixelManagerDelegate) -> Response {
    let user = match auth_user(request, &manager_delegate) {
        Ok(user) => user,
        Err(response) => return response,
    };

    let input = try_or_400!(post_input!(request, {
        pixel_id: String
    }));

    let pixel = match manager_delegate.get_manager().get_pixel_from_encoded(&input.pixel_id, Some(&user)) {
        Ok(pixel) => pixel,
        Err(_) => return Response::text("DB error").with_status_code(500)
    };
    if let Some(pixel) = pixel {
        if manager_delegate.get_manager().delete_pixel(pixel, Some(&user)).is_err() {
            return Response::text("DB error").with_status_code(500);
        }
    } else {
        return Response::empty_400(); // could not find pixel
    }
    // Response::redirect_303("/manage")
    helpers::javascript_redirect("/manage")

}

fn register_hit(pixel_id: &str, request: &Request, manager_delegate: &PixelManagerDelegate) {
    let ua = request.header("User-Agent").unwrap_or_default();
    let ip = request.header("X-Forwarded-For").unwrap_or_default();
    _ = manager_delegate.get_manager().register_hit_with_encoded_pixel(pixel_id, &ip, ua, chrono::offset::Utc::now());
}


fn get_url_for_pixel(pixel: &TrackingPixel) -> String {
    format!("https://track.willem.page/img/{}/white.png", pixel.url_safe_encode())
}
